/* -*- mode: C++; c-file-style: "stroustrup"; c-basic-offset: 4; indent-tabs-mode: nil; -*- */

/* How to compile:
 *
 * gcc -o pmatrix pmatrix.c -O3 -Wall -pthread -DNDEBUG -march=native -fgcse -fgcse-las -fgcse-sm -fopenmp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <math.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/times.h>
#include <assert.h>

unsigned long long mem;
int speedup;
size_t nb_threads;
double base_time;

/* This is used for the random generation of matrices.
 * You can use a constant value if you wish.
 */
#define RANGE(S) (S)

#define BLOCK   55 // Size of the blocks for block-matrix multiplication.

#define BLUE "\033[1;34m"
#define RED  "\033[1;31m"
#define DEF  "\033[0;0m"

void* safe_malloc(size_t size)
{
    void* data = malloc(size);
    if (!data)
    {
        fprintf(stderr, "Could not allocate %u bytes!\n", size);
        abort();
    }
    mem += size;
    return data;
}

typedef void* (*void_f)(void*);

void start_thread(pthread_t *thread, void_f job, void* arg)
{
    static size_t id = 0;
    if (pthread_create(thread, NULL, job, arg))
    {
        fprintf(stderr, "Could not create thread %d, aborting.\n", id);
        abort();
    }
    id++;
}

void join_threads(pthread_t *threads, size_t n)
{
    size_t i;
    for(i = 0; i < n; ++i)
    {
        if (pthread_join(threads[i], NULL))
        {
            fprintf(stderr, "Could not wait for thread %d.\n", i);
            abort();
        }
    }
}

void mutex_init(pthread_mutex_t *mutex)
{
    if (pthread_mutex_init(mutex, NULL))
    {
        fprintf(stderr, "Could not initialize mutex, aborting.\n");
        abort();
    }
}

void cond_init(pthread_cond_t *cond)
{
    if (pthread_cond_init(cond, NULL))
    {
        fprintf(stderr, "Could not initialize condition variable, aborting.\n");
        abort();
    }
}

double* alloc_double(size_t size)
{
    return (double*) safe_malloc(size*sizeof(double));
}

size_t count_cpus()
{
    char s[256];
    size_t count = 0;
    FILE *f = fopen("/proc/stat", "r");
    if (!f)
    {
        fprintf(stderr, "/proc/stat not found!\n");
        abort();
    }
    while(fgets(s,sizeof(s),f) && !strncmp(s,"cpu",3))
    {
        count++;
    }
    fclose(f);
    return count > 1 ? count - 1 : count;
}

typedef void (*mat_f)(double*, double*, double*, size_t);

void timer(struct tms *t, struct timeval *tv)
{
    if (times(t) == (clock_t) -1)
    {
        perror("times");
    }
    if (gettimeofday(tv, NULL) < 0)
    {
        perror("gettimeofday");
    }
}

double timed_call(const char* color, const char *title, mat_f f,
                  double *a, double *b, double *c, size_t dim)
{
    struct tms t1, t2;
    struct timeval v1, v2;
    double u, s, r = 0;
    int low, high;

    fprintf(stdout, "%s%s...", color ? color : DEF, title);
    fflush(stdout);

    if (!color)
    {
        f(a, b, c, dim);
        fprintf(stdout, "done!\n");
    }
    else
    {
        timer(&t1, &v1);
        f(a, b, c, dim);
        timer(&t2, &v2);

        u = t2.tms_utime - t1.tms_utime;
        s = t2.tms_stime - t1.tms_stime;
        low = v2.tv_usec - v1.tv_usec;
        high = v2.tv_sec - v1.tv_sec;
        r = (low / 1000000.0) + high;
        u /= sysconf(_SC_CLK_TCK);
        s /= sysconf(_SC_CLK_TCK);

        if (speedup)
        {
            fprintf(stdout,
                    "done!\treal: %gs, user: %gs, sys: %gs, idle: %g%%, %sS: %g, E: %g, T0: %gs%s\n",
                    r, u, s,
                    r*nb_threads < u+s ? 0.0 : 100*(1.0-(u+s)/(r*nb_threads)),
                    RED,
                    base_time/r,              // S
                    base_time/(r*nb_threads), // E
                    r*nb_threads-base_time,   // T0
                    DEF);
        }
        else
        {
            fprintf(stdout,
                    "done!\treal: %gs, user: %gs, sys: %gs, idle: %g%%%s\n",
                    r, u, s, r < u+s ? 0.0 : 100-(100*(u+s))/r, DEF);
        }
    }

    speedup = 0;
    return r;
}

// Precision for checking identity.
#define PRECISION 1e-6

/* Random generation, but always the same for
 * a given size for fair comparison.
 */
void gen_mat(double *a, double *dummy1, double *dummy2, size_t dim)
{
    size_t i,j;
    srand48(dim);
    for(i = 0; i < dim; ++i)
    {
        for(j = 0; j < dim; ++j)
        {
            double z = (drand48() - 0.5)*RANGE(dim);
            a[i*dim+j] = (z < 10*PRECISION && z > -10*PRECISION) ? 0.0 : z;
        }
    }
}

void row_mat_mult(double *a, double *b, double *c, size_t dim)
{
    // TODO:
    // 1) Copy your previous rearranged matrix multiplication.
    // 2) Parallelize with OpenMP.

}

void block_mat_mult(double *a, double *b, double *c, size_t dim)
{
    // TODO:
    // 1) Copy your previous block matrix multiplication.
    // 2) Parallelise with OpenMP.

}

/* Check that a divider is not too close to 0. */
static void check_divider(double x)
{
    if (fabs(x) < PRECISION)
    {
        fprintf(stderr, "Matrix non-invertible or computationally unstable.\n");
        abort();
    }
}

/* Read a re-arranged matrix. */
#define A(I,J) a[index[I]*n+(J)]
#define B(I,J) b[index[I]*n+(J)]

/* a = invert(input), using b as temporary matrix
 */
void inv_mat(double *input, double *a, double *b, size_t n)
{
    int i,j,k;
    double x;
    size_t index[n];

    // Init a, b, and index.
    memcpy(a, input, n*n*sizeof(double));
    memset(b, 0, n*n*sizeof(double));
    for(i = 0; i < n; ++i)
    {
        index[i] = i;   // Identity mapping.
        b[i*n+i] = 1.0; // Identity matrix.
    }

    // Gaussian elimination.
    for(k = 0; k < n; ++k)
    {
        // Find pivot.
        x = fabs(A(k,k));
        j = k;
        for(i = k+1; i < n; ++i)
        {
            double y = fabs(A(i,k));
            if (y > x)
            {
                j = i;
                x = y;
            }
        }

        // Swap j <-> k.
        i = index[j];
        index[j] = index[k];
        index[k] = i;

        // Division step.
        check_divider(x = A(k,k));
        A(k,k) = 1.0;
        for(j = k+1; j < n; ++j) A(k,j) /= x;
        for(j = 0  ; j < n; ++j) B(k,j) /= x;

        // Elimination step.
        for(i = k+1; i < n; ++i)
        {
            double Aik = A(i,k);
            A(i,k) = 0.0;
            for(j = k+1; j < n; ++j) A(i,j) -= Aik*A(k,j);
            for(j = 0  ; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }

    // Back-substitution.
    for(k = n-1; k > 0; --k)
    {
        for(i = k-1; i >= 0; --i)
        {
            double Aik = A(i,k);
            //A(i,k) = 0.0; -- implicit and not used later.
            for(j = 0; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }

    // Result.
    for(i = 0; i < n; ++i)
    {
        for(j = 0; j < n; ++j)
        {
            a[i*n+j] = B(i,j);
        }
    }
}

/* Read a matrix. */
#undef A
#undef B
#define B(I,J) a[(I)*n+(J)]
#define A(I,J) b[(I)*n+(J)]

void inv_mat_mov(double *input, double *a, double *b, size_t n)
{
    int i,j,k;
    double x;

    // Init a, b, and index.
    memcpy(&A(0,0), input, n*n*sizeof(double));
    memset(&B(0,0), 0, n*n*sizeof(double));
    for(i = 0; i < n; ++i)
    {
        B(i,i) = 1.0; // Identity matrix.
    }

    // Gaussian elimination.
    for(k = 0; k < n; ++k)
    {
        // Find pivot.
        x = fabs(A(k,k));
        j = k;
        for(i = k+1; i < n; ++i)
        {
            double y = fabs(A(i,k));
            if (y > x)
            {
                j = i;
                x = y;
            }
        }

        // Swap j <-> k.
        if (j != k)
        {
            for(i = k; i < n; ++i)
            {
                x = A(j,i);
                A(j,i) = A(k,i);
                A(k,i) = x;
            }
            for(i = 0; i < n; ++i)
            {
                x = B(j,i);
                B(j,i) = B(k,i);
                B(k,i) = x;
            }
        }

        // Division step.
        check_divider(x = A(k,k));
        A(k,k) = 1.0;
        for(j = k+1; j < n; ++j) A(k,j) /= x;
        for(j = 0  ; j < n; ++j) B(k,j) /= x;

        // Elimination step.
        for(i = k+1; i < n; ++i)
        {
            double Aik = A(i,k);
            A(i,k) = 0.0;
            for(j = k+1; j < n; ++j) A(i,j) -= Aik*A(k,j);
            for(j = 0  ; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }

    // Back-substitution.
    for(k = n-1; k > 0; --k)
    {
        for(i = k-1; i >= 0; --i)
        {
            double Aik = A(i,k);
            //A(i,k) = 0.0; -- implicit and not used later.
            for(j = 0; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }
}

typedef struct
{
    double *input, *a, *b;
    size_t n, id, *index;

    pthread_mutex_t *tlock;
    pthread_cond_t *tcond;
    int *tok;
} invmat_data_t;

void set_ok(invmat_data_t *data, int ok)
{
    size_t id = data->id;
    pthread_mutex_t *lock = &data->tlock[id];
    pthread_mutex_lock(lock);
    data->tok[id] = ok;
    pthread_cond_signal(&data->tcond[id]);
    pthread_mutex_unlock(lock);
}

void wait_ok(invmat_data_t *data, int who, int val)
{
    assert(who != data->id || data->tok[who] >= val);
    if (who != data->id)
    {
        pthread_mutex_t *lock = &data->tlock[who];
        pthread_mutex_lock(lock);
        while(data->tok[who] < val)
        {
            pthread_cond_wait(&data->tcond[who], lock);
        }
        pthread_mutex_unlock(lock);
    }
}

// Hook to the global array that you can use.
// Typically memset(gtok, 0, nb_threads*sizeof(int));
int* gtok;

// The barrier & split-barrier.
// 1) Data.

void init_barrier()
{
    // TODO - from assignment 3 + added phase.
}

void barrier()
{
    // TODO - from assignment 3 + added reset.
}

int barrier_arrived()
{
    // TODO - see the book.
    int p;

    return p;
}

void barrier_wait(int p)
{
    // TODO - see the book.
}

void barrier_destroy()
{
    // TODO - destroy the lock & condition variable.
}

typedef void* (*job_f)(invmat_data_t*);

// The kloop condition.
// Data: You'll need a row index (for the k loop)
// + a mutex & condition variable.

void init_kloop(size_t max)
{
    // TODO - initialize the index + mutex + condition variable.
}

void kloop_ok(size_t k)
{
    // TODO - set the the index + broadcast.
}

void kloop_wait(size_t k)
{
    // Wait on the condition variable, careful about what it is.
}

void kloop_destroy()
{
    // TODO - destroy the mutex & condition variable.
}

/* Read a matrix. */
#undef A
#undef B
#define B(I,J) a[(I)*n+(J)]
#define A(I,J) b[(I)*n+(J)]

void* job_inv_mat_mov(invmat_data_t *data)
{
    size_t n = data->n;
    double *a = data->a;
    double *b = data->b;
    int i,j,k;
    double x;

#if 0
    // Useful code if you want to do optional experiments.
    size_t block = n / nb_threads;
    size_t start = data->id * block;
    size_t end = (data->id == nb_threads - 1) ? n : (start + block);
#endif

    // Only one thread does this.
    // The same thread ((k % nb_threads) == data->id matches
    // this one at the 1st iteration) will find the 1st pivot.
    if (data->id == 0)
    {
        // Init a, b, and index.
        memcpy(&A(0,0), data->input, n*n*sizeof(double));
        memset(&B(0,0), 0, n*n*sizeof(double));
        for(i = 0; i < n; ++i)
        {
            B(i,i) = 1.0; // Identity matrix.
        }
    }

    // Gaussian elimination.
    for(k = 0; k < n; ++k)
    {
        // TODO - add parallel code from assignment 3.

        // Find pivot.
        x = fabs(A(k,k));
        j = k;
        for(i = k+1; i < n; ++i)
        {
            double y = fabs(A(i,k));
            if (y > x)
            {
                j = i;
                x = y;
            }
        }
                        
        // Swap j <-> k.
        if (j != k)
        {
            for(i = k; i < n; ++i)
            {
                x = A(j,i);
                A(j,i) = A(k,i);
                A(k,i) = x;
            }
            for(i = 0; i < n; ++i)
            {
                x = B(j,i);
                B(j,i) = B(k,i);
                B(k,i) = x;
            }
        }

        // Division step.
        check_divider(x = A(k,k));
        A(k,k) = 1.0;
        for(j = k+1; j < n; ++j) A(k,j) /= x;
        for(j = 0  ; j < n; ++j) B(k,j) /= x;
        
        
        // Elimination step.
        for(i = k+1; i < n; ++i) if (i % nb_threads == data->id)
        {
            double Aik = A(i,k);
            A(i,k) = 0.0;
            for(j = k+1; j < n; ++j) A(i,j) -= Aik*A(k,j);
            for(j = 0  ; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }


    // Back-substitution.
    for(k = n-1; k > 0; --k)
    {
        for(i = k-1; i >= 0; --i)
        {
            double Aik = A(i,k);
            //A(i,k) = 0.0; -- implicit and not used later.
            for(j = 0; j < n; ++j) B(i,j) -= Aik*B(k,j);
        }
    }

    return NULL;
}

void pthread_inv_mat(double *input, double *a, double *b, size_t n)
{
    invmat_data_t data[nb_threads];
    pthread_cond_t tcond[nb_threads];
    pthread_mutex_t tlock[nb_threads];
    int tok[nb_threads];
    pthread_t threads[nb_threads];
    size_t *index = (size_t*) safe_malloc(n*sizeof(size_t));
    size_t i;
    speedup = 1;
    init_barrier();
    init_kloop(n);
    gtok = tok; // Init hook.
    for(i = 0; i < nb_threads; ++i)
    {
        data[i].input = input;
        data[i].a = a;
        data[i].b = b;
        data[i].n = n;
        data[i].index = index;
        data[i].id = i;
        mutex_init(&tlock[i]);
        cond_init(&tcond[i]);
        tok[i] = 2;
        data[i].tlock = tlock;
        data[i].tcond = tcond;
        data[i].tok = tok;
    }
    for(i = 1; i < nb_threads; ++i)
    {
        start_thread(&threads[i], (void_f)job_inv_mat_mov, &data[i]);
    }
    job_inv_mat_mov(&data[0]);
    join_threads(threads+1, nb_threads-1);
    free(index);
    barrier_destroy();
    kloop_destroy();
}

void check_identity(double *a, double *dummy1, double *dummy2, size_t dim)
{
    size_t i,j;
    double error = 0.0;
    for(i = 0; i < dim; ++i)
    {
        for(j = 0; j < dim; ++j)
        {
            double r = i == j ? a[i*dim+j] - 1.0 : a[i*dim+j];
            if (r < -PRECISION || r > PRECISION)
            {
                fprintf(stderr, "Matrix is not identity.\n");
                abort();
            }
            error += fabs(r);
        }
    }
    /* Smaller error = better precision. */
    fprintf(stdout, "\033[K\rError=%g ", error);
}

void test(size_t dim)
{
    double *a = alloc_double(dim*dim);
    double *b = alloc_double(dim*dim);
    double *c = alloc_double(dim*dim);
    double tinv1, tinv2, tinv;

    timed_call(NULL, "Generating A", gen_mat, a, NULL, NULL, dim);

    tinv1 = timed_call(BLUE, "Inverting", inv_mat, a, b, c, dim);

    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    timed_call(BLUE, "MultiplyingR", row_mat_mult, a, b, c, dim);
    timed_call(NULL, "Checking", check_identity, c, NULL, NULL, dim);

    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    timed_call(BLUE, "MultiplyingB", block_mat_mult, a, b, c, dim);
    timed_call(NULL, "Checking", check_identity, c, NULL, NULL, dim);

    timed_call(NULL, "Randomizing", gen_mat, b, NULL, NULL, dim);
    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    tinv2 = timed_call(BLUE, "InvertingM", inv_mat_mov, a, b, c, dim);
    tinv = tinv1 < tinv2 ? tinv1 : tinv2;

    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    timed_call(BLUE, "MultiplyingB", block_mat_mult, a, b, c, dim);
    timed_call(NULL, "Checking", check_identity, c, NULL, NULL, dim);

    fprintf(stdout, "Using %u threads.\n", nb_threads);

    timed_call(NULL, "Randomizing", gen_mat, b, NULL, NULL, dim);
    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    base_time = tinv;
    timed_call(BLUE, "PInverting", pthread_inv_mat, a, b, c, dim);

    timed_call(NULL, "Randomizing", gen_mat, c, NULL, NULL, dim);
    timed_call(BLUE, "PMultiplyingB", block_mat_mult, a, b, c, dim);
    timed_call(NULL, "Checking", check_identity, c, NULL, NULL, dim);

    free(c);
    free(b);
    free(a);
}

int main(int argc, char *argv[])
{
    if (argc < 2)
    {
        fprintf(stderr, "Usage: %s size [threads]\n", argv[0]);
        return 1;
    }
    else
    {
        int i = atoi(argv[1]);
        int n = argc < 3 ? 0 : atoi(argv[2]);

        if (i < 1)
        {
            fprintf(stderr, "Minimum dimension is 1.\n");
            return 1;
        }

        mem = 0;
        speedup = 0;
        nb_threads = n < 1 ? count_cpus() : n;
        fprintf(stdout, "Threads: %u\n", nb_threads);
        test((size_t) i);
        if (mem < 1 << 10)
        {
            fprintf(stdout, "Used %llu bytes.\n", mem);
        }
        else if (mem < 1 << 20)
        {
            fprintf(stdout, "Used %llu kB.\n", mem >> 10);
        }
        else
        {
            fprintf(stdout, "Used %llu MB.\n", mem >> 20);
        }

        return 0;
    }
}
